pub static NAME: &str = "Fragments";
pub static PKGNAME: &str = "fragments";
pub static APP_ID: &str = "de.haeckerfelix.Fragments";
pub static PATH_ID: &str = "/de/haeckerfelix/Fragments";
pub static VERSION: &str = "3.0.1";
pub static PROFILE: &str = "default";
pub static VCS_TAG: &str = "95bf648";
pub static LOCALEDIR: &str = "/usr/share/locale";
pub static DATADIR: &str = "/usr/share";
